from datetime import date, datetime, timezone
from decimal import Decimal
from ipaddress import IPv4Address, IPv6Address
from typing import (
    Any, Callable, Generic, Iterable, List, Optional, Type, Tuple, TypeVar,
    Union,
)
from uuid import UUID

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo  # type: ignore


PyType = TypeVar('PyType')
JsonType = TypeVar('JsonType', None, int, float, Decimal, str, list)


class BaseType(Generic[PyType, JsonType]):
    __slots__: Tuple[str, ...] = ()

    py_type: Type[PyType]

    @classmethod
    def escape(cls, value: PyType, escape: Callable) -> str:
        return str(value)

    def from_json(self, value: JsonType) -> Optional[PyType]:
        raise NotImplementedError()

    @classmethod
    def to_json(cls, value: PyType, to_json: Callable) -> JsonType:
        raise NotImplementedError()

    def __eq__(self, other):
        return (
            type(self) == type(other) and
            all(
                getattr(self, slot) == getattr(other, slot)
                for slot in self.__slots__
            )
        )

    def __repr__(self):
        return '{}({})'.format(
            type(self).__name__,
            ', '.join(repr(getattr(self, arg)) for arg in self.__slots__)
        )


class StrType(BaseType[str, str]):
    py_type = str

    def __init__(self, *params):
        pass

    @classmethod
    def escape(cls, value: str, escape=None) -> str:
        value = value.replace('\\', '\\\\').replace("'", "\\'")
        return f"'{value}'"

    def from_json(self, value: str) -> str:
        return str(value)

    @classmethod
    def to_json(cls, value: str, to_json: Callable) -> str:
        return value


class StrStripZerosType(StrType):

    def from_json(self, value: str) -> str:
        return value.rstrip('\0')


class IntType(BaseType[int, int]):
    py_type = int

    def from_json(self, value: int) -> int:
        return int(value)

    @classmethod
    def to_json(cls, value: int, to_json: Callable) -> int:
        return value


class FloatType(BaseType[float, float]):
    py_type = float

    def from_json(self, value: float) -> float:
        return float(value)

    @classmethod
    def to_json(cls, value: float, to_json: Callable) -> float:
        return value


class DecimalType(BaseType[Decimal, Decimal]):
    py_type = Decimal

    def __init__(self, *params):
        pass

    @classmethod
    def escape(cls, value: PyType, escape: Callable) -> str:
        return f"'{value}'"

    def from_json(self, value: Decimal) -> Decimal:
        return Decimal(value)

    @classmethod
    def to_json(cls, value: Decimal, to_json: Callable) -> Decimal:
        # Clickhouse requires serializing it without quotes, so we use
        # `simplejson.dumps(..., use_decimal=True)`
        return value


class DateType(BaseType[date, str]):
    py_type = Optional[date]

    @classmethod
    def escape(cls, value: date, escape=None) -> str:
        if value is None:
            return "'0000-00-00'"
        return f"'{value.isoformat()}'"

    @classmethod
    def to_json(cls, value: date, to_json: Callable) -> str:
        return value.isoformat()

    def from_json(self, value: str) -> Optional[date]:
        if value == '0000-00-00':
            return None
        return date.fromisoformat(value)


class DateTimeType(BaseType[datetime, str]):
    py_type = datetime

    __slots__ = ('_tzinfo',)

    def __init__(self, tz_name=None):
        if tz_name is None:
            self._tzinfo = None
        else:
            self._tzinfo = zoneinfo.ZoneInfo(tz_name)

    @classmethod
    def escape(cls, value: datetime, escape=None) -> str:
        value = value.replace(tzinfo=None, microsecond=0)
        return f"'{value.isoformat()}'"

    @classmethod
    def to_json(cls, value: datetime, to_json: Callable) -> str:
        value = value.replace(tzinfo=None, microsecond=0)
        return value.isoformat()

    def from_json(self, value: str) -> Optional[datetime]:
        if value == '0000-00-00 00:00:00':
            return None
        result = datetime.fromisoformat(value)
        if self._tzinfo is not None:
            result = result.replace(tzinfo=self._tzinfo)
        return result


class DateTimeUTCType(DateTimeType):

    @classmethod
    def escape(cls, value: datetime, escape=None) -> str:
        if value.utcoffset() is None:
            raise ValueError(
                'Got naive datetime while timezone-aware is expected'
            )
        value = (
            value.astimezone(timezone.utc)
                .replace(tzinfo=None, microsecond=0)
        )
        return f"'{value.isoformat()}'"

    @classmethod
    def to_json(cls, value: datetime, to_json: Callable) -> str:
        if value.utcoffset() is None:
            raise ValueError(
                'Got naive datetime while timezone-aware is expected'
            )
        value = (
            value.astimezone(timezone.utc)
                .replace(tzinfo=None, microsecond=0)
        )
        return value.isoformat()

    def from_json(self, value: str) -> datetime:
        result = datetime.fromisoformat(value)
        if self._tzinfo is None:
            return result.replace(tzinfo=timezone.utc)
        else:
            return result.replace(tzinfo=self._tzinfo).astimezone(timezone.utc)


class UUIDType(BaseType[UUID, str]):
    py_type = UUID

    @classmethod
    def escape(cls, value: UUID, escape=None) -> str:
        return f"'{value}'"

    @classmethod
    def to_json(cls, value: UUID, to_json: Callable) -> str:
        return str(value)

    def from_json(self, value: str) -> UUID:
        return self.py_type(value)


class IPv4Type(BaseType[IPv4Address, str]):
    py_type = IPv4Address

    @classmethod
    def to_json(cls, value: IPv4Address, to_json: Callable) -> str:
        return str(value)

    def from_json(self, value: str) -> IPv4Address:
        return self.py_type(value)


class IPv6Type(BaseType[IPv6Address, str]):
    py_type = IPv6Address

    @classmethod
    def to_json(cls, value: IPv6Address, to_json: Callable) -> str:
        return str(value)

    def from_json(self, value: str) -> IPv6Address:
        return self.py_type(value)


class NothingType(BaseType[None, None]):
    py_type = Type[None]

    @classmethod
    def escape(cls, value: None, escape=None) -> str:
        return 'NULL'

    @classmethod
    def to_json(cls, value: None, to_json: Callable) -> None:
        return None

    def from_json(self, value: None) -> None:
        # Actually it's never called
        return None # pragma: nocover


class TupleType(BaseType[tuple, list]):
    __slots__ = ('_item_types',)

    py_type = tuple

    def __init__(self, *item_types):
        self._item_types = item_types

    @classmethod
    def escape(cls, value: tuple, escape=None) -> str:
        return '({})'.format(
            ','.join(escape(v) for v in value)
        )

    @classmethod
    def to_json(cls, value: tuple, to_json: Callable) -> List[JsonType]:
        return [to_json(v) for v in value]

    def from_json(self, value: List[JsonType]) -> tuple:
        assert len(self._item_types) == len(value)
        return tuple(
            t.from_json(v) for t, v in zip(self._item_types, value)
        )


class ArrayType(BaseType[list, list]):
    __slots__ = ('_item_type',)

    py_type = list

    def __init__(self, item_type):
        self._item_type = item_type

    @classmethod
    def escape(cls, value: list, escape) -> str:
        return '[{}]'.format(
            ','.join(escape(v) for v in value)
        )

    @classmethod
    def to_json(cls, value: list, to_json: Callable) -> List[JsonType]:
        return [to_json(v) for v in value]

    def from_json(self, value: List[JsonType]) -> list:
        return [self._item_type.from_json(v) for v in value]


class ProxyType(BaseType):

    def __new__(cls, item_type):
        return item_type

    @classmethod
    def escape(cls, value: PyType, escape=None) -> str:
        raise RuntimeError('Must be never called')  # pragma: nocover

    @classmethod
    def to_json(cls, value: PyType, to_json: Callable) -> JsonType:
        raise RuntimeError('Must be never called')  # pragma: nocover


class NullableType(BaseType):
    __slots__ = ('_item_type',)

    def __init__(self, item_type):
        self._item_type = item_type

    @classmethod
    def escape(cls, value: PyType, escape=None) -> str:
        raise RuntimeError('Must be never called')  # pragma: nocover

    @classmethod
    def to_json(cls, value: PyType, to_json: Callable) -> JsonType:
        raise RuntimeError('Must be never called')  # pragma: nocover

    def from_json(self, value: JsonType) -> Any:
        if value is None:
            return
        return self._item_type.from_json(value)


class AggregateFunction:

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'<AggregateFunction (did you forget to finalize aggregation?)>'


class AggregateFunctionType(BaseType):
    __slots__ = ('_item_type',)

    def __init__(self, item_type):
        self._item_type = item_type

    @classmethod
    def to_json(cls, value: PyType, to_json: Callable) -> JsonType:
        raise RuntimeError('Must be never called')  # pragma: nocover

    def from_json(self, value: JsonType) -> AggregateFunction:
        return AggregateFunction(value)


DEFAULT_CONVERTES = [
    (StrType, ['String', 'Enum8', 'Enum16'], str),
    (StrStripZerosType, ['FixedString']),
    (
        IntType,
        [
            'UInt8', 'UInt16', 'UInt32', 'UInt64',
            'Int8', 'Int16', 'Int32', 'Int64',
        ],
        [int, bool],
    ),
    (FloatType, ['Float32', 'Float64'], float),
    # It looks like Decimal32/Decimal64/Decimal128 were used for some period
    # (reproducable with Clickhouse 20.1.2.4)
    (DecimalType, ['Decimal', 'Decimal32', 'Decimal64', 'Decimal128'], Decimal),
    (DateType, ['Date'], date),
    (DateTimeType, ['DateTime'], datetime),
    (UUIDType, ['UUID'], UUID),
    (IPv4Type, ['IPv4'], IPv4Address),
    (IPv6Type, ['IPv6'], IPv6Address),
    (NothingType, ['Nothing'], type(None)),
    (TupleType, ['Tuple'], tuple),
    (ArrayType, ['Array'], list),
    (NullableType, ['Nullable']),
    (ProxyType, ['LowCardinality', 'SimpleAggregateFunction']),
    (AggregateFunctionType, ['AggregateFunction']),
]


class TypeRegistry:

    def __init__(self, converters=DEFAULT_CONVERTES):
        self._types = {}
        self._escapers = {}
        self._to_json = {}
        for args in converters:
            self.register(*args)

    def register(
        self,
        conv_class: Type[BaseType],
        ch_types: Iterable[str],
        py_types: Union[type, Iterable[type]] = (),
    ):
        for ch_type in ch_types:
            self._types[ch_type] = conv_class
        if isinstance(py_types, type):
            py_types = [py_types]
        for py_type in py_types:
            assert isinstance(py_type, type)
            self._escapers[py_type] = conv_class.escape
            self._to_json[py_type] = conv_class.to_json

    def __getitem__(self, ch_type_name):
        return self._types[ch_type_name]

    def escape(self, value):
        py_type = type(value)
        try:
            return self._escapers[py_type](value, self.escape)
        except KeyError:
            # Fallback to slower method
            for subclass in py_type.mro()[1:]:
                if subclass in self._escapers:
                    escape = self._escapers[subclass]
                    # Cache to speed up further look-ups
                    self._escapers[py_type] = escape
                    return escape(value, self.escape)
            else:
                raise TypeError(f'Unsupported type {py_type}')

    def to_json(self, value):
        py_type = type(value)
        try:
            return self._to_json[py_type](value, self.to_json)
        except KeyError:
            # Fallback to slower method
            for subclass in py_type.mro()[1:]:
                if subclass in self._to_json:
                    to_json = self._to_json[subclass]
                    # Cache to speed up further look-ups
                    self._to_json[py_type] = to_json
                    return to_json(value, self.to_json)
            else:
                raise TypeError(f'Unsupported type {py_type}')



from typing import Optional, Union
from urllib.parse import parse_qsl, urlsplit, urlunsplit

from aiohttp.client import ClientSession, ClientTimeout

from .client import Client


def dsn_to_params(dsn):
    parsed = urlsplit(dsn)

    if parsed.scheme != 'clickhouse':
        raise ValueError(
            f'Expecting "clickhouse" scheme in DSN, got {parsed.scheme}'
        )

    hostname = parsed.hostname or '127.0.0.1'
    port = parsed.port or 8123
    netloc = f'{hostname}:{port}'

    database = parsed.path.lstrip('/')
    if not database:
        database = 'default'

    params = dict(parse_qsl(parsed.query))

    return {
        **params,
        'url': urlunsplit(('http', netloc, '', '', '')),
        'database': database,
        'user': parsed.username,
        'password': parsed.password,
    }


class Pool:

    DEFAULT_TIMEOUT = {
        'total': 5*60,
        'connect': None,
        'sock_read': None,
        'sock_connect': 10,
    }

    def __init__(
        self, dsn, session_class=ClientSession,
        session_timeout: Optional[Union[float, int, dict]] = None,
        client_class=Client, **params,
    ):
        timeout_params = self.DEFAULT_TIMEOUT.copy()
        if isinstance(session_timeout, dict):
            timeout_params.update(session_timeout)
        else:
            timeout_params['total'] = session_timeout  # type: ignore
        self._session = session_class(
            timeout=ClientTimeout(**timeout_params)
        )
        params.update(dsn_to_params(dsn))
        self._client = client_class(self._session, **params)

    async def close(self):
        await self._session.close()

    def __await__(self):
        # For compartibility with asyncpg (`await create_pool(...)`)
        yield from []
        return self

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def acquire(self, *, timeout=None):
        return self._client

    async def release(self, conn, *, timeout=None):
        pass

    async def iterate(self, *args, **kwargs):
        async for row in self._client.iterate(*args, **kwargs):
            yield row

    async def execute(self, *args, **kwargs):
        return await self._client.execute(*args, **kwargs)

    async def fetch(self, *args, **kwargs):
        return await self._client.fetch(*args, **kwargs)

    async def fetchrow(self, *args, **kwargs):
        return await self._client.fetchrow(*args, **kwargs)

    async def fetchval(self, *args, **kwargs):
        return await self._client.fetchval(*args, **kwargs)


def connect(dsn, **kwargs):
    return Pool(dsn, **kwargs)

def create_pool(dsn, **kwargs):
    return Pool(dsn, **kwargs)



from clickhouse_sqlalchemy.drivers.base import ClickHouseCompiler
from clickhouse_sqlalchemy.drivers.http.base import ClickHouseDialect_http
from sqlalchemy import exc
from sqlalchemy.sql import crud


class ClickhouseSaSQLCompiler(ClickHouseCompiler):

    _clickhouse_json_each_row = False

    def get_from_hint_text(self, table, text):
        return text

    def visit_insert(self, insert_stmt, asfrom=False, **kw):
        assert not self.stack # INSERT only at top level

        self.stack.append(
            {
                "correlate_froms": set(),
                "asfrom_froms": set(),
                "selectable": insert_stmt,
            }
        )

        preparer = self.preparer

        text = "INSERT INTO "
        table_text = preparer.format_table(insert_stmt.table)
        text += table_text

        if insert_stmt.select is not None:
            # `_setup_crud_params()` multiplies parameter placeholders for
            # multiparam inserts.  We don't want this, so this part is moved to
            # the branch for `INSERT INTO ... SELECT`.
            crud_params = crud._setup_crud_params(
                self, insert_stmt, crud.ISINSERT, **kw
            )

            if not crud_params: # pragma: no cover
                raise exc.CompileError(
                    "The '%s' dialect with current database "
                    "version settings does not support empty "
                    "inserts." % self.dialect.name
                )

            assert not insert_stmt._has_multi_parameters

            text += " (%s)" % ", ".join(
                [preparer.format_column(c[0]) for c in crud_params]
            )

            select_text = self.process(self._insert_from_select, **kw)

            # TODO Provide visit_cte for Clickhouse variant of CTE
            if self.ctes:
                text += " %s%s" % (self._render_cte_clause(), select_text)
            else:
                text += " %s" % select_text
        else:
            # This is normally done by `crud._setup_crud_params()`
            self.isinsert = True

            self._clickhouse_json_each_row = True
            text += ' FORMAT JSONEachRow'

        assert insert_stmt._post_values_clause is None

        self.stack.pop(-1)

        assert not asfrom
        return text


class ClickhouseSaDialect(ClickHouseDialect_http):
    statement_compiler = ClickhouseSaSQLCompiler

    supports_empty_insert = False
    cte_follows_insert = True



import asyncio
from datetime import datetime
from decimal import Decimal
import os

import clickhouse_sqlalchemy
import pytest
import sqlalchemy as sa

import aiochsa
from aiochsa import error_codes
from aiochsa.dialect import ClickhouseSaDialect


def pytest_collection_modifyitems(items):
    for item in items:
        if asyncio.iscoroutinefunction(item.function):
            item.add_marker('asyncio')


def pytest_addoption(parser):
    parser.addoption(
        '--clickhouse-version', default=None,
        help='use specified version of ClickHouse',
    )


@pytest.fixture(scope='session')
def dsn(request, docker_services):
    ch_version = request.config.getoption('--clickhouse-version')
    if ch_version is not None:
        os.environ['CLICKHOUSE_VERSION'] = ch_version
    docker_services.start('clickhouse-server')
    public_port = docker_services.wait_for_service('clickhouse-server', 8123)
    return f'clickhouse://{docker_services.docker_ip}:{public_port}'


_clickhouse_version = None


@pytest.fixture
async def clickhouse_version(dsn):
    global _clickhouse_version
    if _clickhouse_version is None:
        async with aiochsa.connect(dsn) as conn:
            version = await conn.fetchval('SELECT version()')
            _clickhouse_version = tuple(
                int(num) for num in version.split('.')
            )
    return _clickhouse_version


@pytest.fixture
async def pool(dsn):
    async with aiochsa.create_pool(dsn) as pool:
        yield pool


@pytest.fixture
async def conn(dsn):
    async with aiochsa.connect(dsn) as conn:
        yield conn


@pytest.fixture
def recreate_table(conn):

    async def _recreate(table_name, create_ddl):
        try:
            await conn.execute(f'DROP TABLE {table_name}')
        except aiochsa.DBException as exc:
            if exc.code != error_codes.UNKNOWN_TABLE:
                raise

        await conn.execute(create_ddl)

    return _recreate


TEST1_CREATE_DDL = '''\
CREATE TABLE test
(
    id UInt64,
    enum Enum8(''=0, 'ONE'=1, 'TWO'=2),
    name String,
    timestamp DateTime,
    amount Decimal64(9)
)
ENGINE = MergeTree()
ORDER BY id
'''


@pytest.fixture
async def table_test(recreate_table):
    await recreate_table('test', TEST1_CREATE_DDL)

    return sa.Table(
        'test', sa.MetaData(),
        sa.Column('id', sa.Integer),
        sa.Column('enum', sa.Enum, default=''),
        sa.Column('name', sa.String),
        sa.Column('timestamp', sa.DateTime, default=datetime.utcnow),
        sa.Column('amount', sa.DECIMAL, default=Decimal(0)),
    )


TEST_MT_CREATE_DDL = '''\
CREATE TABLE test_mt
(
    title String,
    num UInt64
)
ENGINE = MergeTree()
ORDER BY num
'''


@pytest.fixture
async def table_mt(recreate_table):
    await recreate_table('test_mt', TEST_MT_CREATE_DDL)

    return sa.Table(
        'test_mt', sa.MetaData(),
        sa.Column('num', sa.Integer),
        sa.Column('title', sa.String),
    )


TEST_SMT_CREATE_DDL = '''\
CREATE TABLE table_smt
(
    key UInt8,
    value UInt64
)
ENGINE = SummingMergeTree()
ORDER BY key
'''


@pytest.fixture
async def table_smt(recreate_table):
    await recreate_table('table_smt', TEST_SMT_CREATE_DDL)

    return sa.Table(
        'table_smt', sa.MetaData(),
        sa.Column('key', sa.Integer),
        sa.Column('value', sa.Integer),
    )


TEST_AMT_CREATE_DDL = '''\
CREATE TABLE test_amt
(
    key UInt8,
    value AggregateFunction(sum, UInt8)
)
ENGINE = AggregatingMergeTree()
ORDER BY key
'''


@pytest.fixture
async def table_amt(recreate_table):
    await recreate_table('test_amt', TEST_AMT_CREATE_DDL)

    return sa.Table(
        'test_amt', sa.MetaData(),
        sa.Column('key', sa.Integer),
        sa.Column('value', sa.Integer),
    )


TABLE_FOR_TYPE_DDL_TEMPLATE = '''\
CREATE TABLE test_for_type
(
    value {type}
)
ENGINE = Log()
'''


@pytest.fixture
def recreate_table_for_type(recreate_table):

    async def _recreate(type_str):
        create_ddl = TABLE_FOR_TYPE_DDL_TEMPLATE.format(type=type_str)
        await recreate_table('test_for_type', create_ddl)
        return 'test_for_type'

    return _recreate


@pytest.fixture
def table_for_type(recreate_table_for_type):

    async def _create(sa_type):
        if isinstance(sa_type, type):
            sa_type = sa_type()
        type_str = sa_type.compile(dialect=ClickhouseSaDialect())

        table_name = await recreate_table_for_type(type_str)

        return sa.Table(
            table_name, sa.MetaData(),
            sa.Column('value', sa_type),
        )

    return _create


@pytest.fixture(params=[
    sa.select,
    clickhouse_sqlalchemy.select,
    aiochsa.select,
])
def any_select(request):
    return request.param



import pytest

from aiochsa.record import Record


@pytest.fixture
def record():
    return Record(['a', 'b'], [1, 'v'])


def test_len(record):
    assert len(record) == 2


def test_getitem_name(record):
    assert record['a'] == 1
    assert record['b'] == 'v'
    with pytest.raises(KeyError):
        record['c']


def test_getitem_index(record):
    assert record[0] == record[-2] == 1
    assert record[1] == record[-1] == 'v'
    with pytest.raises(IndexError):
        record[2]
    with pytest.raises(IndexError):
        record[-3]


def test_slice(record):
    assert record[:] == record[:2] == [1, 'v']
    assert record[2:] == record[:-2] == []
    assert record[1:] == record[1:2] == ['v']
    assert record[:1] == record[:-1] == [1]


def test_get(record):
    assert record.get('a') == record.get(0) == 1
    assert record.get('b') == record.get(1) == 'v'
    assert record.get('c') is record.get(2) is None
    default = object()
    assert record.get('c', default) is record.get(2, default) is default


def test_to_sequence(record):
    assert tuple(record) == (1, 'v')
    assert list(record) == [1, 'v']
    assert list(iter(record)) == [1, 'v']
    assert [*record] == [1, 'v']


def test_to_dict(record):
    assert dict(record) == {'a': 1, 'b': 'v'}
    assert {**record} == {'a': 1, 'b': 'v'}


def test_eq_dict(record):
    assert record == {'a': 1, 'b': 'v'}
    assert record == {'b': 'v', 'a': 1}
    assert record != {'a': 2, 'b': 'v'}
    assert record != {'a': 1, 'c': 'v'}
    assert record != {'a': 1}
    assert record != {'a': 1, 'b': 'v', 'c': 2}


@pytest.mark.parametrize('t', [tuple, list, iter])
def test_eq_sequence(record, t):
    assert record == t([1, 'v'])
    assert record != t(['v', 1])
    assert record != t([1])
    assert record != t([1, 'v', 2])


def test_mapping_interface(record):
    assert list(record.keys()) == ['a', 'b']
    assert list(record.values()) == [1, 'v']
    assert list(record.items()) == [('a', 1), ('b', 'v')]


def test_repr(record):
    assert repr(record) == "<Record a=1 b='v'>"



from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
import enum
from ipaddress import IPv4Address, IPv6Address
import itertools
from random import randrange
from typing import Iterable, Tuple, Type, Union
import uuid

import pytest
import sqlalchemy as sa
from clickhouse_sqlalchemy import types as t

import aiochsa
from aiochsa.types import (
    ArrayType, BaseType, DateTimeUTCType, IntType, NullableType, ProxyType,
    StrType, TupleType, TypeRegistry,
)


def parametrized_id(value):
    if isinstance(value, type):
        # Types without parameters, e.g. `Int32` -> `Int32()`
        value = value()
    return repr(value)


MAX_IPV4 = 256**4 - 1
MAX_IPV6 = 256**16 - 1


class PyEnum(str, enum.Enum):
    FOO = 'FOO'
    BAR = 'BAR'
    BAZ = 'BAZ'


# Int version of enum is needed for definition in CAST only.

class CHEnum8(int, enum.Enum):
    FOO = -128
    BAR = 0
    BAZ = 127

class CHEnum16(int, enum.Enum):
    FOO = -32768
    BAR = 0
    BAZ = 32767


class CustomStr(str):
    pass


SaType = Union[sa.types.TypeEngine, Type[sa.types.TypeEngine]]


def combine_typed_rapameters(spec_seq: Iterable[Tuple[SaType, Iterable]]):
    return list( # Wrap into list to make it reusable (iterator is one-off)
        itertools.chain(*[
            list(itertools.product([sa_type], values))
            for sa_type, values in spec_seq
        ])
    )


TYPED_PARAMETERS = combine_typed_rapameters([
    (t.String, ['', '\0', "'", '"', '\\', "\\'", 'зразок', CustomStr('abc')]),
    (t.String(16), ['', 'зразок']),

    (t.Enum8(CHEnum8), list(PyEnum)),
    (t.Enum16(CHEnum16), list(PyEnum)),

    (t.Int8, [-128, 0, 127]),
    (t.UInt8, [0, 255]),
    (t.Int16, [-32768, 0, 32767]),
    (t.UInt16, [0, 65535]),
    (t.Int32, [-2147483648, 0, 2147483647]),
    (t.UInt32, [0, 4294967295]),
    (t.Int64, [-9223372036854775808, 0, 9223372036854775807]),
    (t.UInt64, [0, 18446744073709551615]),

    # `nan`, `inf`, `-inf` are returned as `null` in JSON formats
    # 0. and -0. are considered equal in Python
    (t.Float32, [0., -0., 1e-37, -1e-37, 1e38, -1e38]),
    (t.Float64, [0., -0., 1e-307, -1e-307, 1e308, -1e308]),

    (t.Decimal(9, 4), [Decimal(), Decimal('1.2345'), Decimal('-1.2345')]),
    (t.Decimal(18, 9), [
        Decimal(0), Decimal('1.234567891'), Decimal('-1.234567891'),
    ]),
    (t.Decimal(38, 19), [
        Decimal(0), Decimal('1.234567891234567891'),
        Decimal('-1.234567891234567891'),
    ]),

    # Start of the epoch is not supported by Clickhouse, at least one
    # second or day must be added
    (t.Date, [date(1970, 1, 2), date.today()]),
    (t.DateTime, [
        datetime(1970, 1, 1, 0, 0, 1),
        datetime.now().replace(microsecond=0)
    ]),

    (t.UUID, [uuid.uuid4()]),
    (t.IPv4, map(IPv4Address, [0, MAX_IPV4, randrange(1, MAX_IPV4)])),
    (t.IPv6, map(IPv6Address, [0, MAX_IPV6, randrange(1, MAX_IPV6)])),

    (t.Nullable(t.String), [None, '', 'abc']),
    (t.LowCardinality(t.String), ['', 'abc']),
    (t.LowCardinality(t.Nullable(t.String)), [None, '', 'abc']),
    (t.Array(t.String), [['foo', 'bar']]),

    # TODO Tests for `Tuple`, including deeply nested
])


@pytest.mark.parametrize(
    'sa_type,value',
    TYPED_PARAMETERS,
    ids = parametrized_id,
)
async def test_cast_round(conn, sa_type, value):
    # Select parameters go through `escape()` method of type
    result = await conn.fetchval(
        sa.select([sa.func.cast(value, sa_type)])
    )
    assert result == value


@pytest.mark.parametrize(
    'sa_type,value',
    TYPED_PARAMETERS,
    ids = parametrized_id,
)
async def test_insert_select_round(conn, table_for_type, sa_type, value):
    # Insert parameters go through `to_json()` method of type
    table = await table_for_type(sa_type)
    await conn.execute(
        table.insert(),
        {'value': value},
    )
    result = await conn.fetchval(
        table.select()
    )
    assert result == value


@pytest.mark.parametrize(
    'value',
    [
        '', '\0', "'", '"', '\\', "\\'", 'зразок',
        -9223372036854775808, 0, 9223372036854775807,
        0., -0., 1e-307, -1e-307, 1e308, -1e308,
        [1, 2, 3], ['a', 'b', 'c'],
        None,
        ('abc', 123, 2.34, None, [1, 2, 3], (1, ['a', 'b', 'c'])),
        [('abc', 123), ('bcd', 234)],
    ],
    ids = parametrized_id,
)
async def test_as_is_round(conn, value):
    result = await conn.fetchval(
        # Without `bindparam()` the value is handled by SQLAlchemy as column
        # expression and not passed through escape
        sa.select([sa.bindparam('a')])
            .params(a=value)
    )
    assert result == value


@pytest.mark.parametrize(
    'sa_type,value',
    [
        (t.Date, '0000-00-00'),
        (t.DateTime, '0000-00-00 00:00:00'),
    ],
    ids = parametrized_id,
)
async def test_zero_dates(clickhouse_version, conn, sa_type, value):
    if clickhouse_version >= (20, 7):
        pytest.skip('Feature is dropped in 20.7')
    result = await conn.fetchval(
        sa.select([sa.func.cast(value, sa_type)])
    )
    assert result is None


@pytest.mark.parametrize('tz_name,tz_offset', [
    ('UTC', 0),
    ('EST', -18_000),
    ('Europe/Moscow', 10_800),
])
async def test_timezones(conn, tz_name, tz_offset):
    dt = datetime(2020, 1, 1)
    result = await conn.fetchval(
        sa.func.toTimeZone(sa.func.toDateTime(dt), tz_name).select()
    )
    assert result.utcoffset().total_seconds() == tz_offset
    assert result.astimezone(timezone.utc).replace(tzinfo=None) == dt


@pytest.fixture
async def conn_utc(dsn):
    types = TypeRegistry()
    types.register(DateTimeUTCType, ['DateTime'], datetime)
    async with aiochsa.connect(dsn, types=types) as conn:
        yield conn


DATETIME_UTC_PARAMETERS = [
    datetime(1970, 1, 1, 0, 0, 1, tzinfo=timezone.utc),
    datetime.utcnow().replace(tzinfo=timezone.utc, microsecond=0),
]


@pytest.mark.parametrize(
    'value',
    DATETIME_UTC_PARAMETERS,
    ids = parametrized_id,
)
async def test_datetime_utc_escape(conn_utc, value):
    # Select parameters go through `escape()` method of type
    result = await conn_utc.fetchval(
        sa.func.toDateTime(value)
    )
    assert result == value


@pytest.mark.parametrize(
    'value',
    DATETIME_UTC_PARAMETERS,
    ids = parametrized_id,
)
async def test_datetime_utc_insert_select(conn_utc, table_for_type, value):
    # Insert parameters go through `to_json()` method of type
    table = await table_for_type(sa.DateTime)
    await conn_utc.execute(
        table.insert(),
        {'value': value},
    )

    result = await conn_utc.fetchval(
        table.select()
    )
    assert result == value


async def test_datetime_utc_escape_naive(conn_utc):
    # Select parameters go through `escape()` method of type
    with pytest.raises(ValueError):
        await conn_utc.fetchval(
            sa.func.toDateTime(datetime.now())
        )


async def test_datetime_utc_insert_naive(conn_utc, table_for_type):
    # Insert parameters go through `to_json()` method of type
    table = await table_for_type(sa.DateTime)
    with pytest.raises(ValueError):
        await conn_utc.execute(
            table.insert(),
            {'value': datetime.now()},
        )


@pytest.mark.parametrize('tz_name,tz_offset', [
    ('UTC', 0),
    ('EST', -18_000),
    ('Europe/Moscow', 10_800),
])
async def test_timezones_with_utc(conn_utc, tz_name, tz_offset):
    dt = datetime(2020, 1, 1, tzinfo=timezone.utc)
    result = await conn_utc.fetchval(
        sa.func.toTimeZone(sa.func.toDateTime(dt, 'UTC'), tz_name).select()
    )
    assert result == dt


@pytest.mark.parametrize('value', [0, 4294967295])
async def test_simple_aggregate_function(conn, recreate_table_for_type, value):
    table_name = await recreate_table_for_type(
        'SimpleAggregateFunction(max, "UInt32")'
    )
    await conn.execute(f'INSERT INTO {table_name} VALUES ({value})')
    result = await conn.fetchval(f'SELECT value FROM {table_name}')
    assert result == value


# Parser tests would be meaningless if something is wrong with `__eq__`, so
# it's better to insure it works.

def _get_all_subclasses(cls):
    for subcls in cls.__subclasses__():
        yield subcls
        yield from _get_all_subclasses(subcls)

SIMPLE_TYPE_CLASSES = [
    cls for cls in _get_all_subclasses(BaseType)
    if not cls.__slots__ and not issubclass(cls, ProxyType)
]


@pytest.mark.parametrize(
    'type_class', SIMPLE_TYPE_CLASSES,
)
def test_eq_simple(type_class):
    type_obj = type_class()
    assert type_obj == type_class()
    assert all(
        type_obj != other_type_class()
        for other_type_class in SIMPLE_TYPE_CLASSES
        if other_type_class is not type_class
    )


@pytest.mark.parametrize(
    'type_class',
    [ArrayType, NullableType],
)
def test_eq_wrapping(type_class):
    type_obj = type_class(StrType())
    assert type_obj == type_class(StrType())

    assert type_obj != StrType()
    assert repr(type_obj) != repr(StrType())

    assert type_obj != type_class(IntType())
    assert repr(type_obj) != repr(type_class(IntType()))


def test_eq_tuple():
    type_obj = TupleType(StrType(), IntType())
    assert type_obj == TupleType(StrType(), IntType())

    assert type_obj != TupleType(StrType())
    assert repr(type_obj) != repr(TupleType(StrType()))



import asyncio
from decimal import Decimal

import pytest

from aiochsa.parser import parse_type, parse_json_compact
from aiochsa import types as t


@pytest.mark.parametrize(
    'type_str,type_obj',
    [
        ('String', t.StrType()),
        ("Enum8('' = -128, 'a' = 0, '\t\n\0\\\'' = 127)", t.StrType()),
        ("Enum16('a' = 1)", t.StrType()),
        ('FixedString(8)', t.StrStripZerosType()),
        *[
            (ts, t.IntType()) for ts in [
                'UInt8', 'UInt16', 'UInt32', 'UInt64',
                'Int8', 'Int16', 'Int32', 'Int64',
            ]
        ],
        *[
            (ts, t.FloatType()) for ts in ['Float32', 'Float64']
        ],
        ('Decimal(18, 9)', t.DecimalType()),
        ('DateTime', t.DateTimeType()),
        ('Date', t.DateType()),
        ('UUID', t.UUIDType()),
        ('IPv4', t.IPv4Type()),
        ('IPv6', t.IPv6Type()),
        ('Nothing', t.NothingType()),
        ('Nullable(Nothing)', t.NullableType(t.NothingType())),
        ('LowCardinality(String)', t.StrType()),
        ('LowCardinality(Nullable(String))', t.NullableType(t.StrType())),
        ('SimpleAggregateFunction(max, UInt32)', t.IntType()),
        ('Tuple(Int8, String)', t.TupleType(t.IntType(), t.StrType())),
        ('Array(String)', t.ArrayType(t.StrType())),
        (
            'Tuple('
                'UInt8, '
                'Array('
                    'Nullable(FixedString(2))'
                '), '
                'Array('
                    "Tuple(Decimal(9, 2), Enum16('' = -128))"
                '), '
                'Tuple('
                    "Nullable(Enum8('' = -128, 'a' = 1, '\t\n\0\\\'' = 2)), "
                    'Nullable(Nothing)'
                ')'
            ')',
            t.TupleType(
                t.IntType(),
                t.ArrayType(
                    t.NullableType(t.StrStripZerosType())
                ),
                t.ArrayType(
                    t.TupleType(t.DecimalType(), t.StrType()),
                ),
                t.TupleType(
                    t.NullableType(t.StrType()),
                    t.NullableType(t.NothingType()),
                ),
            )
        ),
    ],
)
def test_parse_type(type_str, type_obj):
    types = t.TypeRegistry()
    result = parse_type(types, type_str)
    assert result == type_obj


async def test_parse_json_compact_decimal():
    """ Insure Decimal is parsed without loss of precision """
    content = b'''\
        {
            "meta": [
                {"name": "column", "type": "Decimal(18, 13)"}
            ],
            "data": [
                [1.2345678901230]
            ]
        }
    '''
    [[value]] = list(parse_json_compact(t.TypeRegistry(), content))
    assert isinstance(value, Decimal)
    assert str(value) == '1.2345678901230'



from decimal import Decimal

import pytest

import aiochsa
from aiochsa import error_codes


async def test_exc(conn):
    with pytest.raises(aiochsa.DBException) as exc_info:
        await conn.execute('ERROR')

    assert exc_info.value.code == error_codes.SYNTAX_ERROR
    assert 'Syntax error' in exc_info.value.display_text

    assert str(error_codes.SYNTAX_ERROR) in str(exc_info.value)
    assert 'Syntax error' in str(exc_info.value)


@pytest.mark.parametrize(
    'statement', [
        'ERROR',
        pytest.param('ERROR ' * 1000, id='ERROR ...'),
    ],
)
async def test_exc_stacktrace(dsn, statement):
    async with aiochsa.connect(dsn, stacktrace=1) as conn:
        with pytest.raises(aiochsa.DBException) as exc_info:
            await conn.execute(statement)

    assert exc_info.value.code == error_codes.SYNTAX_ERROR
    assert 'Syntax error' in exc_info.value.display_text
    assert exc_info.value.stack_trace is not None

    assert str(error_codes.SYNTAX_ERROR) in str(exc_info.value)
    assert 'Syntax error' in str(exc_info.value)
    assert len(str(exc_info.value)) < len(exc_info.value.display_text) + 250


async def test_exc_row(conn, table_test, clickhouse_version):
    with pytest.raises(aiochsa.DBException) as exc_info:
        await conn.execute(
            table_test.insert(),
            {'amount': Decimal('1234567890.1234567890')},
        )
    assert exc_info.value.code == error_codes.ARGUMENT_OUT_OF_BOUND
    if clickhouse_version >= (20, 3, 9, 70):
        assert exc_info.value.row == (1, '{"amount": 1234567890.1234567890}')



# Source:
# https://github.com/ClickHouse/ClickHouse/blob/master/src/Common/ErrorCodes.cpp

OK = 0
UNSUPPORTED_METHOD = 1
UNSUPPORTED_PARAMETER = 2
UNEXPECTED_END_OF_FILE = 3
EXPECTED_END_OF_FILE = 4
CANNOT_PARSE_TEXT = 6
INCORRECT_NUMBER_OF_COLUMNS = 7
THERE_IS_NO_COLUMN = 8
SIZES_OF_COLUMNS_DOESNT_MATCH = 9
NOT_FOUND_COLUMN_IN_BLOCK = 10
POSITION_OUT_OF_BOUND = 11
PARAMETER_OUT_OF_BOUND = 12
SIZES_OF_COLUMNS_IN_TUPLE_DOESNT_MATCH = 13
DUPLICATE_COLUMN = 15
NO_SUCH_COLUMN_IN_TABLE = 16
DELIMITER_IN_STRING_LITERAL_DOESNT_MATCH = 17
CANNOT_INSERT_ELEMENT_INTO_CONSTANT_COLUMN = 18
SIZE_OF_FIXED_STRING_DOESNT_MATCH = 19
NUMBER_OF_COLUMNS_DOESNT_MATCH = 20
CANNOT_READ_ALL_DATA_FROM_TAB_SEPARATED_INPUT = 21
CANNOT_PARSE_ALL_VALUE_FROM_TAB_SEPARATED_INPUT = 22
CANNOT_READ_FROM_ISTREAM = 23
CANNOT_WRITE_TO_OSTREAM = 24
CANNOT_PARSE_ESCAPE_SEQUENCE = 25
CANNOT_PARSE_QUOTED_STRING = 26
CANNOT_PARSE_INPUT_ASSERTION_FAILED = 27
CANNOT_PRINT_FLOAT_OR_DOUBLE_NUMBER = 28
CANNOT_PRINT_INTEGER = 29
CANNOT_READ_SIZE_OF_COMPRESSED_CHUNK = 30
CANNOT_READ_COMPRESSED_CHUNK = 31
ATTEMPT_TO_READ_AFTER_EOF = 32
CANNOT_READ_ALL_DATA = 33
TOO_MANY_ARGUMENTS_FOR_FUNCTION = 34
TOO_FEW_ARGUMENTS_FOR_FUNCTION = 35
BAD_ARGUMENTS = 36
UNKNOWN_ELEMENT_IN_AST = 37
CANNOT_PARSE_DATE = 38
TOO_LARGE_SIZE_COMPRESSED = 39
CHECKSUM_DOESNT_MATCH = 40
CANNOT_PARSE_DATETIME = 41
NUMBER_OF_ARGUMENTS_DOESNT_MATCH = 42
ILLEGAL_TYPE_OF_ARGUMENT = 43
ILLEGAL_COLUMN = 44
ILLEGAL_NUMBER_OF_RESULT_COLUMNS = 45
UNKNOWN_FUNCTION = 46
UNKNOWN_IDENTIFIER = 47
NOT_IMPLEMENTED = 48
LOGICAL_ERROR = 49
UNKNOWN_TYPE = 50
EMPTY_LIST_OF_COLUMNS_QUERIED = 51
COLUMN_QUERIED_MORE_THAN_ONCE = 52
TYPE_MISMATCH = 53
STORAGE_DOESNT_ALLOW_PARAMETERS = 54
STORAGE_REQUIRES_PARAMETER = 55
UNKNOWN_STORAGE = 56
TABLE_ALREADY_EXISTS = 57
TABLE_METADATA_ALREADY_EXISTS = 58
ILLEGAL_TYPE_OF_COLUMN_FOR_FILTER = 59
UNKNOWN_TABLE = 60
ONLY_FILTER_COLUMN_IN_BLOCK = 61
SYNTAX_ERROR = 62
UNKNOWN_AGGREGATE_FUNCTION = 63
CANNOT_READ_AGGREGATE_FUNCTION_FROM_TEXT = 64
CANNOT_WRITE_AGGREGATE_FUNCTION_AS_TEXT = 65
NOT_A_COLUMN = 66
ILLEGAL_KEY_OF_AGGREGATION = 67
CANNOT_GET_SIZE_OF_FIELD = 68
ARGUMENT_OUT_OF_BOUND = 69
CANNOT_CONVERT_TYPE = 70
CANNOT_WRITE_AFTER_END_OF_BUFFER = 71
CANNOT_PARSE_NUMBER = 72
UNKNOWN_FORMAT = 73
CANNOT_READ_FROM_FILE_DESCRIPTOR = 74
CANNOT_WRITE_TO_FILE_DESCRIPTOR = 75
CANNOT_OPEN_FILE = 76
CANNOT_CLOSE_FILE = 77
UNKNOWN_TYPE_OF_QUERY = 78
INCORRECT_FILE_NAME = 79
INCORRECT_QUERY = 80
UNKNOWN_DATABASE = 81
DATABASE_ALREADY_EXISTS = 82
DIRECTORY_DOESNT_EXIST = 83
DIRECTORY_ALREADY_EXISTS = 84
FORMAT_IS_NOT_SUITABLE_FOR_INPUT = 85
RECEIVED_ERROR_FROM_REMOTE_IO_SERVER = 86
CANNOT_SEEK_THROUGH_FILE = 87
CANNOT_TRUNCATE_FILE = 88
UNKNOWN_COMPRESSION_METHOD = 89
EMPTY_LIST_OF_COLUMNS_PASSED = 90
SIZES_OF_MARKS_FILES_ARE_INCONSISTENT = 91
EMPTY_DATA_PASSED = 92
UNKNOWN_AGGREGATED_DATA_VARIANT = 93
CANNOT_MERGE_DIFFERENT_AGGREGATED_DATA_VARIANTS = 94
CANNOT_READ_FROM_SOCKET = 95
CANNOT_WRITE_TO_SOCKET = 96
CANNOT_READ_ALL_DATA_FROM_CHUNKED_INPUT = 97
CANNOT_WRITE_TO_EMPTY_BLOCK_OUTPUT_STREAM = 98
UNKNOWN_PACKET_FROM_CLIENT = 99
UNKNOWN_PACKET_FROM_SERVER = 100
UNEXPECTED_PACKET_FROM_CLIENT = 101
UNEXPECTED_PACKET_FROM_SERVER = 102
RECEIVED_DATA_FOR_WRONG_QUERY_ID = 103
TOO_SMALL_BUFFER_SIZE = 104
CANNOT_READ_HISTORY = 105
CANNOT_APPEND_HISTORY = 106
FILE_DOESNT_EXIST = 107
NO_DATA_TO_INSERT = 108
CANNOT_BLOCK_SIGNAL = 109
CANNOT_UNBLOCK_SIGNAL = 110
CANNOT_MANIPULATE_SIGSET = 111
CANNOT_WAIT_FOR_SIGNAL = 112
THERE_IS_NO_SESSION = 113
CANNOT_CLOCK_GETTIME = 114
UNKNOWN_SETTING = 115
THERE_IS_NO_DEFAULT_VALUE = 116
INCORRECT_DATA = 117
ENGINE_REQUIRED = 119
CANNOT_INSERT_VALUE_OF_DIFFERENT_SIZE_INTO_TUPLE = 120
UNSUPPORTED_JOIN_KEYS = 121
INCOMPATIBLE_COLUMNS = 122
UNKNOWN_TYPE_OF_AST_NODE = 123
INCORRECT_ELEMENT_OF_SET = 124
INCORRECT_RESULT_OF_SCALAR_SUBQUERY = 125
CANNOT_GET_RETURN_TYPE = 126
ILLEGAL_INDEX = 127
TOO_LARGE_ARRAY_SIZE = 128
FUNCTION_IS_SPECIAL = 129
CANNOT_READ_ARRAY_FROM_TEXT = 130
TOO_LARGE_STRING_SIZE = 131
CANNOT_CREATE_TABLE_FROM_METADATA = 132  # removed
AGGREGATE_FUNCTION_DOESNT_ALLOW_PARAMETERS = 133
PARAMETERS_TO_AGGREGATE_FUNCTIONS_MUST_BE_LITERALS = 134
ZERO_ARRAY_OR_TUPLE_INDEX = 135
UNKNOWN_ELEMENT_IN_CONFIG = 137
EXCESSIVE_ELEMENT_IN_CONFIG = 138
NO_ELEMENTS_IN_CONFIG = 139
ALL_REQUESTED_COLUMNS_ARE_MISSING = 140
SAMPLING_NOT_SUPPORTED = 141
NOT_FOUND_NODE = 142
FOUND_MORE_THAN_ONE_NODE = 143
FIRST_DATE_IS_BIGGER_THAN_LAST_DATE = 144
UNKNOWN_OVERFLOW_MODE = 145
QUERY_SECTION_DOESNT_MAKE_SENSE = 146
NOT_FOUND_FUNCTION_ELEMENT_FOR_AGGREGATE = 147
NOT_FOUND_RELATION_ELEMENT_FOR_CONDITION = 148
NOT_FOUND_RHS_ELEMENT_FOR_CONDITION = 149
EMPTY_LIST_OF_ATTRIBUTES_PASSED = 150
INDEX_OF_COLUMN_IN_SORT_CLAUSE_IS_OUT_OF_RANGE = 151
UNKNOWN_DIRECTION_OF_SORTING = 152
ILLEGAL_DIVISION = 153
AGGREGATE_FUNCTION_NOT_APPLICABLE = 154
UNKNOWN_RELATION = 155
DICTIONARIES_WAS_NOT_LOADED = 156
ILLEGAL_OVERFLOW_MODE = 157
TOO_MANY_ROWS = 158
TIMEOUT_EXCEEDED = 159
TOO_SLOW = 160
TOO_MANY_COLUMNS = 161
TOO_DEEP_SUBQUERIES = 162
TOO_DEEP_PIPELINE = 163
READONLY = 164
TOO_MANY_TEMPORARY_COLUMNS = 165
TOO_MANY_TEMPORARY_NON_CONST_COLUMNS = 166
TOO_DEEP_AST = 167
TOO_BIG_AST = 168
BAD_TYPE_OF_FIELD = 169
BAD_GET = 170
BLOCKS_HAVE_DIFFERENT_STRUCTURE = 171  # removed
CANNOT_CREATE_DIRECTORY = 172
CANNOT_ALLOCATE_MEMORY = 173
CYCLIC_ALIASES = 174
CHUNK_NOT_FOUND = 176
DUPLICATE_CHUNK_NAME = 177
MULTIPLE_ALIASES_FOR_EXPRESSION = 178
MULTIPLE_EXPRESSIONS_FOR_ALIAS = 179
THERE_IS_NO_PROFILE = 180
ILLEGAL_FINAL = 181
ILLEGAL_PREWHERE = 182
UNEXPECTED_EXPRESSION = 183
ILLEGAL_AGGREGATION = 184
UNSUPPORTED_MYISAM_BLOCK_TYPE = 185
UNSUPPORTED_COLLATION_LOCALE = 186
COLLATION_COMPARISON_FAILED = 187
UNKNOWN_ACTION = 188
TABLE_MUST_NOT_BE_CREATED_MANUALLY = 189
SIZES_OF_ARRAYS_DOESNT_MATCH = 190
SET_SIZE_LIMIT_EXCEEDED = 191
UNKNOWN_USER = 192
WRONG_PASSWORD = 193
REQUIRED_PASSWORD = 194
IP_ADDRESS_NOT_ALLOWED = 195
UNKNOWN_ADDRESS_PATTERN_TYPE = 196
SERVER_REVISION_IS_TOO_OLD = 197
DNS_ERROR = 198
UNKNOWN_QUOTA = 199
QUOTA_DOESNT_ALLOW_KEYS = 200
QUOTA_EXPIRED = 201
TOO_MANY_SIMULTANEOUS_QUERIES = 202
NO_FREE_CONNECTION = 203
CANNOT_FSYNC = 204
NESTED_TYPE_TOO_DEEP = 205
ALIAS_REQUIRED = 206
AMBIGUOUS_IDENTIFIER = 207
EMPTY_NESTED_TABLE = 208
SOCKET_TIMEOUT = 209
NETWORK_ERROR = 210
EMPTY_QUERY = 211
UNKNOWN_LOAD_BALANCING = 212
UNKNOWN_TOTALS_MODE = 213
CANNOT_STATVFS = 214
NOT_AN_AGGREGATE = 215
QUERY_WITH_SAME_ID_IS_ALREADY_RUNNING = 216
CLIENT_HAS_CONNECTED_TO_WRONG_PORT = 217
TABLE_IS_DROPPED = 218
DATABASE_NOT_EMPTY = 219
DUPLICATE_INTERSERVER_IO_ENDPOINT = 220
NO_SUCH_INTERSERVER_IO_ENDPOINT = 221
ADDING_REPLICA_TO_NON_EMPTY_TABLE = 222
UNEXPECTED_AST_STRUCTURE = 223
REPLICA_IS_ALREADY_ACTIVE = 224
NO_ZOOKEEPER = 225
NO_FILE_IN_DATA_PART = 226
UNEXPECTED_FILE_IN_DATA_PART = 227
BAD_SIZE_OF_FILE_IN_DATA_PART = 228
QUERY_IS_TOO_LARGE = 229
NOT_FOUND_EXPECTED_DATA_PART = 230
TOO_MANY_UNEXPECTED_DATA_PARTS = 231
NO_SUCH_DATA_PART = 232
BAD_DATA_PART_NAME = 233
NO_REPLICA_HAS_PART = 234
DUPLICATE_DATA_PART = 235
ABORTED = 236
NO_REPLICA_NAME_GIVEN = 237
FORMAT_VERSION_TOO_OLD = 238
CANNOT_MUNMAP = 239
CANNOT_MREMAP = 240
MEMORY_LIMIT_EXCEEDED = 241
TABLE_IS_READ_ONLY = 242
NOT_ENOUGH_SPACE = 243
UNEXPECTED_ZOOKEEPER_ERROR = 244
CORRUPTED_DATA = 246
INCORRECT_MARK = 247
INVALID_PARTITION_VALUE = 248
NOT_ENOUGH_BLOCK_NUMBERS = 250
NO_SUCH_REPLICA = 251
TOO_MANY_PARTS = 252
REPLICA_IS_ALREADY_EXIST = 253
NO_ACTIVE_REPLICAS = 254
TOO_MANY_RETRIES_TO_FETCH_PARTS = 255
PARTITION_ALREADY_EXISTS = 256
PARTITION_DOESNT_EXIST = 257
UNION_ALL_RESULT_STRUCTURES_MISMATCH = 258
CLIENT_OUTPUT_FORMAT_SPECIFIED = 260
UNKNOWN_BLOCK_INFO_FIELD = 261
BAD_COLLATION = 262
CANNOT_COMPILE_CODE = 263
INCOMPATIBLE_TYPE_OF_JOIN = 264
NO_AVAILABLE_REPLICA = 265
MISMATCH_REPLICAS_DATA_SOURCES = 266
STORAGE_DOESNT_SUPPORT_PARALLEL_REPLICAS = 267
CPUID_ERROR = 268
INFINITE_LOOP = 269
CANNOT_COMPRESS = 270
CANNOT_DECOMPRESS = 271
CANNOT_IO_SUBMIT = 272
CANNOT_IO_GETEVENTS = 273
AIO_READ_ERROR = 274
AIO_WRITE_ERROR = 275
INDEX_NOT_USED = 277
LEADERSHIP_LOST = 278  # removed
ALL_CONNECTION_TRIES_FAILED = 279
NO_AVAILABLE_DATA = 280
DICTIONARY_IS_EMPTY = 281
INCORRECT_INDEX = 282
UNKNOWN_DISTRIBUTED_PRODUCT_MODE = 283
UNKNOWN_GLOBAL_SUBQUERIES_METHOD = 284  # old name
WRONG_GLOBAL_SUBQUERY = 284
TOO_FEW_LIVE_REPLICAS = 285
UNSATISFIED_QUORUM_FOR_PREVIOUS_WRITE = 286
UNKNOWN_FORMAT_VERSION = 287
DISTRIBUTED_IN_JOIN_SUBQUERY_DENIED = 288
REPLICA_IS_NOT_IN_QUORUM = 289
LIMIT_EXCEEDED = 290
DATABASE_ACCESS_DENIED = 291
LEADERSHIP_CHANGED = 292  # removed
MONGODB_CANNOT_AUTHENTICATE = 293
INVALID_BLOCK_EXTRA_INFO = 294
RECEIVED_EMPTY_DATA = 295
NO_REMOTE_SHARD_FOUND = 296
SHARD_HAS_NO_CONNECTIONS = 297
CANNOT_PIPE = 298
CANNOT_FORK = 299
CANNOT_DLSYM = 300
CANNOT_CREATE_CHILD_PROCESS = 301
CHILD_WAS_NOT_EXITED_NORMALLY = 302
CANNOT_SELECT = 303
CANNOT_WAITPID = 304
TABLE_WAS_NOT_DROPPED = 305
TOO_DEEP_RECURSION = 306
TOO_MANY_BYTES = 307
UNEXPECTED_NODE_IN_ZOOKEEPER = 308
FUNCTION_CANNOT_HAVE_PARAMETERS = 309
INVALID_SHARD_WEIGHT = 317
INVALID_CONFIG_PARAMETER = 318
UNKNOWN_STATUS_OF_INSERT = 319
VALUE_IS_OUT_OF_RANGE_OF_DATA_TYPE = 321
BARRIER_TIMEOUT = 335
UNKNOWN_DATABASE_ENGINE = 336
DDL_GUARD_IS_ACTIVE = 337
UNFINISHED = 341
METADATA_MISMATCH = 342
SUPPORT_IS_DISABLED = 344
TABLE_DIFFERS_TOO_MUCH = 345
CANNOT_CONVERT_CHARSET = 346
CANNOT_LOAD_CONFIG = 347
CANNOT_INSERT_NULL_IN_ORDINARY_COLUMN = 349
INCOMPATIBLE_SOURCE_TABLES = 350
AMBIGUOUS_TABLE_NAME = 351
AMBIGUOUS_COLUMN_NAME = 352
INDEX_OF_POSITIONAL_ARGUMENT_IS_OUT_OF_RANGE = 353
ZLIB_INFLATE_FAILED = 354
ZLIB_DEFLATE_FAILED = 355
BAD_LAMBDA = 356
RESERVED_IDENTIFIER_NAME = 357
INTO_OUTFILE_NOT_ALLOWED = 358
TABLE_SIZE_EXCEEDS_MAX_DROP_SIZE_LIMIT = 359
CANNOT_CREATE_CHARSET_CONVERTER = 360
SEEK_POSITION_OUT_OF_BOUND = 361
CURRENT_WRITE_BUFFER_IS_EXHAUSTED = 362
CANNOT_CREATE_IO_BUFFER = 363
RECEIVED_ERROR_TOO_MANY_REQUESTS = 364
OUTPUT_IS_NOT_SORTED = 365  # removed
SIZES_OF_NESTED_COLUMNS_ARE_INCONSISTENT = 366
TOO_MANY_FETCHES = 367
BAD_CAST = 368  # removed
ALL_REPLICAS_ARE_STALE = 369
DATA_TYPE_CANNOT_BE_USED_IN_TABLES = 370
INCONSISTENT_CLUSTER_DEFINITION = 371
SESSION_NOT_FOUND = 372
SESSION_IS_LOCKED = 373
INVALID_SESSION_TIMEOUT = 374
CANNOT_DLOPEN = 375
CANNOT_PARSE_UUID = 376
ILLEGAL_SYNTAX_FOR_DATA_TYPE = 377
DATA_TYPE_CANNOT_HAVE_ARGUMENTS = 378
UNKNOWN_STATUS_OF_DISTRIBUTED_DDL_TASK = 379
CANNOT_KILL = 380
HTTP_LENGTH_REQUIRED = 381
CANNOT_LOAD_CATBOOST_MODEL = 382
CANNOT_APPLY_CATBOOST_MODEL = 383
PART_IS_TEMPORARILY_LOCKED = 384
MULTIPLE_STREAMS_REQUIRED = 385
NO_COMMON_TYPE = 386
DICTIONARY_ALREADY_EXISTS = 387
CANNOT_ASSIGN_OPTIMIZE = 388
INSERT_WAS_DEDUPLICATED = 389
CANNOT_GET_CREATE_TABLE_QUERY = 390
EXTERNAL_LIBRARY_ERROR = 391
QUERY_IS_PROHIBITED = 392
THERE_IS_NO_QUERY = 393
QUERY_WAS_CANCELLED = 394
FUNCTION_THROW_IF_VALUE_IS_NON_ZERO = 395
TOO_MANY_ROWS_OR_BYTES = 396
QUERY_IS_NOT_SUPPORTED_IN_MATERIALIZED_VIEW = 397
UNKNOWN_MUTATION_COMMAND = 398
FORMAT_IS_NOT_SUITABLE_FOR_OUTPUT = 399
CANNOT_STAT = 400
FEATURE_IS_NOT_ENABLED_AT_BUILD_TIME = 401
CANNOT_IOSETUP = 402
INVALID_JOIN_ON_EXPRESSION = 403
BAD_ODBC_CONNECTION_STRING = 404
PARTITION_SIZE_EXCEEDS_MAX_DROP_SIZE_LIMIT = 405
TOP_AND_LIMIT_TOGETHER = 406
DECIMAL_OVERFLOW = 407
BAD_REQUEST_PARAMETER = 408
EXTERNAL_EXECUTABLE_NOT_FOUND = 409
EXTERNAL_SERVER_IS_NOT_RESPONDING = 410
PTHREAD_ERROR = 411
NETLINK_ERROR = 412
CANNOT_SET_SIGNAL_HANDLER = 413
CANNOT_READLINE = 414  # removed
ALL_REPLICAS_LOST = 415
REPLICA_STATUS_CHANGED = 416
EXPECTED_ALL_OR_ANY = 417
UNKNOWN_JOIN = 418
MULTIPLE_ASSIGNMENTS_TO_COLUMN = 419
CANNOT_UPDATE_COLUMN = 420
CANNOT_ADD_DIFFERENT_AGGREGATE_STATES = 421
UNSUPPORTED_URI_SCHEME = 422
CANNOT_GETTIMEOFDAY = 423
CANNOT_LINK = 424
SYSTEM_ERROR = 425
NULL_POINTER_DEREFERENCE = 426  # removed
CANNOT_COMPILE_REGEXP = 427
UNKNOWN_LOG_LEVEL = 428
FAILED_TO_GETPWUID = 429
MISMATCHING_USERS_FOR_PROCESS_AND_DATA = 430
ILLEGAL_SYNTAX_FOR_CODEC_TYPE = 431
UNKNOWN_CODEC = 432
ILLEGAL_CODEC_PARAMETER = 433
CANNOT_PARSE_PROTOBUF_SCHEMA = 434
NO_DATA_FOR_REQUIRED_PROTOBUF_FIELD = 435  # old name
NO_COLUMN_SERIALIZED_TO_REQUIRED_PROTOBUF_FIELD = 435
PROTOBUF_BAD_CAST = 436
PROTOBUF_FIELD_NOT_REPEATED = 437
DATA_TYPE_CANNOT_BE_PROMOTED = 438
CANNOT_SCHEDULE_TASK = 439
INVALID_LIMIT_EXPRESSION = 440
CANNOT_PARSE_DOMAIN_VALUE_FROM_STRING = 441
BAD_DATABASE_FOR_TEMPORARY_TABLE = 442
NO_COMMON_COLUMNS_WITH_PROTOBUF_SCHEMA = 443  # old name
NO_COLUMNS_SERIALIZED_TO_PROTOBUF_FIELDS = 443
UNKNOWN_PROTOBUF_FORMAT = 444
CANNOT_MPROTECT = 445
FUNCTION_NOT_ALLOWED = 446
HYPERSCAN_CANNOT_SCAN_TEXT = 447
BROTLI_READ_FAILED = 448
BROTLI_WRITE_FAILED = 449
BAD_TTL_EXPRESSION = 450
BAD_TTL_FILE = 451
SETTING_CONSTRAINT_VIOLATION = 452
MYSQL_CLIENT_INSUFFICIENT_CAPABILITIES = 453
OPENSSL_ERROR = 454
SUSPICIOUS_TYPE_FOR_LOW_CARDINALITY = 455
UNKNOWN_QUERY_PARAMETER = 456
BAD_QUERY_PARAMETER = 457
CANNOT_UNLINK = 458
CANNOT_SET_THREAD_PRIORITY = 459
CANNOT_CREATE_TIMER = 460
CANNOT_SET_TIMER_PERIOD = 461
CANNOT_DELETE_TIMER = 462
CANNOT_FCNTL = 463
CANNOT_PARSE_ELF = 464
CANNOT_PARSE_DWARF = 465
INSECURE_PATH = 466
CANNOT_PARSE_BOOL = 467
CANNOT_PTHREAD_ATTR = 468
VIOLATED_CONSTRAINT = 469
QUERY_IS_NOT_SUPPORTED_IN_LIVE_VIEW = 470
INVALID_SETTING_VALUE = 471
READONLY_SETTING = 472
DEADLOCK_AVOIDED = 473
INVALID_TEMPLATE_FORMAT = 474
INVALID_WITH_FILL_EXPRESSION = 475
WITH_TIES_WITHOUT_ORDER_BY = 476
INVALID_USAGE_OF_INPUT = 477
UNKNOWN_POLICY = 478
UNKNOWN_DISK = 479
UNKNOWN_PROTOCOL = 480
PATH_ACCESS_DENIED = 481
DICTIONARY_ACCESS_DENIED = 482
TOO_MANY_REDIRECTS = 483
INTERNAL_REDIS_ERROR = 484
SCALAR_ALREADY_EXISTS = 485
CANNOT_GET_CREATE_DICTIONARY_QUERY = 487
UNKNOWN_DICTIONARY = 488
INCORRECT_DICTIONARY_DEFINITION = 489
CANNOT_FORMAT_DATETIME = 490
UNACCEPTABLE_URL = 491
ACCESS_ENTITY_NOT_FOUND = 492
ACCESS_ENTITY_ALREADY_EXISTS = 493
ACCESS_ENTITY_FOUND_DUPLICATES = 494
ACCESS_STORAGE_READONLY = 495
QUOTA_REQUIRES_CLIENT_KEY = 496
ACCESS_DENIED = 497
LIMIT_BY_WITH_TIES_IS_NOT_SUPPORTED = 498
S3_ERROR = 499
CANNOT_CREATE_DATABASE = 501
CANNOT_SIGQUEUE = 502
AGGREGATE_FUNCTION_THROW = 503
FILE_ALREADY_EXISTS = 504
CANNOT_DELETE_DIRECTORY = 505
UNEXPECTED_ERROR_CODE = 506
UNABLE_TO_SKIP_UNUSED_SHARDS = 507
UNKNOWN_ACCESS_TYPE = 508
INVALID_GRANT = 509
CACHE_DICTIONARY_UPDATE_FAIL = 510
UNKNOWN_ROLE = 511
SET_NON_GRANTED_ROLE = 512
UNKNOWN_PART_TYPE = 513
ACCESS_STORAGE_FOR_INSERTION_NOT_FOUND = 514
INCORRECT_ACCESS_ENTITY_DEFINITION = 515
AUTHENTICATION_FAILED = 516
CANNOT_ASSIGN_ALTER = 517
CANNOT_COMMIT_OFFSET = 518
NO_REMOTE_SHARD_AVAILABLE = 519
CANNOT_DETACH_DICTIONARY_AS_TABLE = 520
ATOMIC_RENAME_FAIL = 521
OPENCL_ERROR = 522  # removed
UNKNOWN_ROW_POLICY = 523
ALTER_OF_COLUMN_IS_FORBIDDEN = 524
INCORRECT_DISK_INDEX = 525
UNKNOWN_VOLUME_TYPE = 526
NO_SUITABLE_FUNCTION_IMPLEMENTATION = 527
CASSANDRA_INTERNAL_ERROR = 528
NOT_A_LEADER = 529
CANNOT_CONNECT_RABBITMQ = 530
CANNOT_FSTAT = 531
LDAP_ERROR = 532
INCONSISTENT_RESERVATIONS = 533
NO_RESERVATIONS_PROVIDED = 534
UNKNOWN_RAID_TYPE = 535
CANNOT_RESTORE_FROM_FIELD_DUMP = 536
ILLEGAL_MYSQL_VARIABLE = 537
MYSQL_SYNTAX_ERROR = 538
CANNOT_BIND_RABBITMQ_EXCHANGE = 539
CANNOT_DECLARE_RABBITMQ_EXCHANGE = 540
CANNOT_CREATE_RABBITMQ_QUEUE_BINDING = 541
CANNOT_REMOVE_RABBITMQ_EXCHANGE = 542
UNKNOWN_MYSQL_DATATYPES_SUPPORT_LEVEL = 543
ROW_AND_ROWS_TOGETHER = 544
FIRST_AND_NEXT_TOGETHER = 545
NO_ROW_DELIMITER = 546
INVALID_RAID_TYPE = 547
UNKNOWN_VOLUME = 548
DATA_TYPE_CANNOT_BE_USED_IN_KEY = 549
CONDITIONAL_TREE_PARENT_NOT_FOUND = 550
ILLEGAL_PROJECTION_MANIPULATOR = 551
UNRECOGNIZED_ARGUMENTS = 552
LZMA_STREAM_ENCODER_FAILED = 553
LZMA_STREAM_DECODER_FAILED = 554
ROCKSDB_ERROR = 555
SYNC_MYSQL_USER_ACCESS_ERROR = 556
UNKNOWN_UNION = 557
EXPECTED_ALL_OR_DISTINCT = 558
INVALID_GRPC_QUERY_INFO = 559
ZSTD_ENCODER_FAILED = 560
ZSTD_DECODER_FAILED = 561
TLD_LIST_NOT_FOUND = 562
CANNOT_READ_MAP_FROM_TEXT = 563
INTERSERVER_SCHEME_DOESNT_MATCH = 564
TOO_MANY_PARTITIONS = 565
CANNOT_RMDIR = 566
DUPLICATED_PART_UUIDS = 567
RAFT_ERROR = 568
MULTIPLE_COLUMNS_SERIALIZED_TO_SAME_PROTOBUF_FIELD = 569
DATA_TYPE_INCOMPATIBLE_WITH_PROTOBUF_FIELD = 570

KEEPER_EXCEPTION = 999
POCO_EXCEPTION = 1000
STD_EXCEPTION = 1001
UNKNOWN_EXCEPTION = 1002
METRIKA_OTHER_ERROR = 1003  # removed
INVALID_SHARD_ID = 1003

CONDITIONAL_TREE_PARENT_NOT_FOUND = 2001  # removed
ILLEGAL_PROJECTION_MANIPULATOR = 2002  # removed
UNRECOGNIZED_ARGUMENTS = 2003  # removed



from collections import namedtuple
import pkgutil
import simplejson as json
from typing import Iterable

from lark import Lark, Transformer, v_args

from .record import Record
from .types import TypeRegistry


__all__ = ['parse_type', 'parse_json_compact', 'JSONDecodeError']


# Re-export
JSONDecodeError = json.JSONDecodeError


type_parser = Lark(
    pkgutil.get_data(__name__, 'type.lark').decode(),  # type: ignore
    parser='lalr',
)


EnumOption = namedtuple('EnumOption', ['label', 'value'])


@v_args(inline=True)
class TypeTransformer(Transformer):

    def __init__(self, types: TypeRegistry):
        self._types = types

    def start(self, type_):
        return type_

    def composite_type(self, name, *types):
        return self._types[name](*types)

    def aggregate_type(self, name, func, type_):
        return self._types[name](type_)

    def simple_type(self, name, *params):
        return self._types[name](*params)

    def enum_param(self, label, value):
        return EnumOption(label, value)

    def STRING(self, value):
        assert value[0] == value[-1] == "'"
        return value[1:-1]

    def INT(self, value):
        return int(value)


def parse_type(types: TypeRegistry, type_str):
    tree = type_parser.parse(type_str)
    return TypeTransformer(types).transform(tree)


def parse_json_compact(
    types: TypeRegistry, content: bytes,
) -> Iterable[Record]:
    # The method is split into three phases:
    #   1. Parse JSON.  It's done immediately, so that we can fall back to
    #      parsing exception when it breaks normal response.
    #   2. Parse type information from meta.  It's done at first iteration of
    #      returned generator.
    #   3. Convert each row one-by-one.  It's done on demand at each iteration.
    # This way we can cleanup resources even when result is not used.

    json_data = json.loads(content, parse_float=str)
    return convert_json_compact(types, json_data)


def convert_json_compact(
    types: TypeRegistry, json_data: dict,
) -> Iterable[Record]:
    names = []
    converters = []
    for column in json_data['meta']:
        names.append(column['name'])
        type_obj = parse_type(types, column['type'])
        converters.append(type_obj.from_json)

    for row in json_data['data']:
        yield Record(
            names = names,
            values = [
                converter(value)
                for converter, value in zip(converters, row)
            ]
        )
